import numpy as np
import numpy.typing as npt


def shuffle_cards() -> npt.NDArray:
    suites = np.array([x for x in range(4)])
    numbers = np.array([x for x in range(13)])
    deck = np.array(np.meshgrid(suites, numbers)).T.reshape(-1, 2)
    np.random.shuffle(deck)
    return deck

def pick_card(n: int, deck: npt.NDArray) -> str:
    display_suite = {0:"Clubs", 1:"Spades", 2:"Hearts", 3:"Diamonds"}
    display_number = {0:"Ace"} | {v:str(v+1) for v in range(1,10)} | {10:"Jack", 11:"Queen", 12:"King"}
    return display_number[deck[n][1]] + " of " + display_suite[deck[n][0]]

def get_code() -> list[str]:
    with open("menutesting.py", "r") as code:
        lines = code.readlines()
    code.close()
    return lines


from menucmd.dsl import build_menus
from menucmd.tests.testing_utils import *
import os


def main():
    path = os.path.join(os.path.dirname(__file__), 'dsltesting.mcmd')
    menus = build_menus(path)

    menus['main_menu']()

#################################################################################################################

if __name__ == "__main__":
    main()



#==================================================================================
#Result Type

class Result():
    def __init__(self, n: int, expand: bool = False):
        self.__n__ = n
        self.__attr__ = None
        self.__expand__ = expand

    def __getitem__(self, n: int):
        r = type(self)(n, expand=self.__expand__)
        r.__attr__ = self.__attr__
        return r

    def __repr__(self):
        return f"<class Result[{self.__n__}]>" + (".expand()" if self.__expand__ else "")

    def __eq__(self, other):
        """Compare by attribute if both self and other have attributes"""
        if not isinstance(other, Result) or self.__expand__ != other.__expand__:
            return False

        if self.__attr__ is not None and other.__attr__ is not None:
            return self.__attr__ == other.__attr__
        else:
            return self.__n__ == other.__n__

    def expand(self):
        """Type to replace and expand in place.   *result <=> result.expand()"""
        r = type(self)(self.__n__, expand=True)
        r.__attr__ = self.__attr__
        return r

    def __getattr__(self, tag):
        """Index by attribute first.  If no attribute, index by index"""
        new_result = type(self)(self.__n__, expand=self.__expand__)
        new_result.__attr__ = tag
        return new_result



import re



def lines_to_dict(lines: list[str]) -> dict:
    """
    Generic recursive descent parser for DSL format.
    
    Grammar (indentation-based):
    - Lines ending with ':' are block headers (e.g., Menu:, Item:, Colors:, func:)
    - Other lines are key:value pairs
    - Indentation (multiples of 4 spaces) determines nesting depth
    - Comments (#) and blank lines are stripped
    
    Returns a generic nested dict where:
    - Block types are keys (e.g., 'Menu', 'Item', 'Colors')
    - Multiple blocks of same type become lists
    - Non-opinionated: no semantic assumptions about block types
    - Fully extensible: add new block types by just using 'BlockName:'
    
    Example output structure:
    {
        "Menu": [
            {
                "name": "Hub",
                "id": "main_menu",
                "Colors": {...},
                "ExitColors": {...},
                "Item": [
                    {"key": "e", "message": "Test", "func": ["func(args)", ...], "Colors": {...}},
                    {"key": "b", "message": "Another", ...}
                ]
            },
            {
                "name": "Lazy Eval",
                "id": "lazy_menu",
                ...
            }
        ]
    }
    """
    lines = [line for line in lines if line.strip() and not line.strip().startswith('#')]
    if not lines:
        return {}
    result, _ = _parse_block(lines, 0, 0)
    return result


def _parse_block(lines: list[str], start_idx: int, expected_indent: int, last_block_name: str = '_Top') -> tuple[dict, int]:
    """
    Recursively parse all lines at a given indent level.
    
    Returns:
        (dict of parsed content, next unprocessed line index)
    """
    block = {}
    n = start_idx
    
    while n < len(lines):
        line = lines[n]
        current_indent = _get_indent(line)
        stripped = line.strip()
        
        # Exit Scope
        if current_indent < expected_indent:
            return block, n
        
        # Over-Indent
        if current_indent > expected_indent:
            raise IndentationError(line)
        
        # Block header (any line ending with ':')
        if stripped.endswith(':'):
            block_name = stripped[:-1]

            # First block must be Menu
            #if block_name != 'Menu' and n == 0:
            #    raise SyntaxError(f"{block_name} must be a descendent of Menu")
            
            # Menu Heirarchy rules
            error = False
            match block_name:
                case 'Menu':
                    error = last_block_name in { 'Colors', 'Item', 'ExitColors', 'Menu' }
                case 'Item':
                    error = last_block_name in { 'Colors', 'Item', 'ExitColors', "_Top" } 
                case 'Colors':
                    error = last_block_name in { 'Colors', 'ExitColors' }
                case 'ExitColors':
                    error = last_block_name in { 'Colors', 'Item', 'ExitColors' }

            if error:
                raise SyntaxError(f"{block_name} cannot be a descendent of {last_block_name}.")

            # Go to next stack
            sub_block, n = _parse_block(lines, n + 1, expected_indent + 1, block_name)
            
            # Add allowed duplicate BLOCKS here
            if block_name in { "Menu", "Item" }:
                if not block_name in block:
                    block[block_name] = []
                block[block_name].append(sub_block)
            else:
                if not block_name in block:
                    block[block_name] = sub_block
                else:
                    raise AttributeError(f"Duplicate block: {block_name}")
        
        # Key:value pair
        elif ':' in stripped:
            key, value = _parse_kv_line(stripped, last_block_name)
            
            # Add allowed duplicate KEYS here
            if key in { "func" }:
                if not key in block:
                    block[key] = []
                block[key].append(value)
            else:
                if not key in block:
                    block[key] = value
                else:
                    raise AttributeError(f"Duplicate attribute: {key}")
            n += 1
        else:
            raise SyntaxError(f"Unrecognized format: \n{stripped}")

    return block, n


def check_quotes(key: str, val: str, block_name: str):
    """Enforce quotes for string values."""

    # Keys that are expected to be quoted
    quoted_keys = { 'exit_key', 'exit_message' } \
    | ({ 'key', 'message', 'empty_message' } if block_name in { "Item" } else set()) \
    | ({ 'name', 'empty_message', 'invalid_key' } if block_name in { "Menu" } else set())

    if key in quoted_keys:
        if not ((val.startswith('"') and val.endswith('"')) or
                (val.startswith("'") and val.endswith("'"))):
            raise TypeError(
                f"Value '{val}' must be a quoted string. "
                f'Try "{val}"'
            )


def _parse_kv_line(line: str, block_name: str) -> tuple[str, str]:
    """
    Parse a key:value line.
    
    Handles quoted strings (both single and double quotes).
    Returns (key, value) where value is unquoted if it was quoted.
    """
    #split_pattern = r"([^:]+):([^:]+.*)"
    match = re.match(r'([^:]+):\s*(.*)', line)
    if not match:
        raise ValueError(f"Invalid key:value syntax: {line}")
    
    key = match.group(1).strip()
    value = match.group(2).strip()
    
    # Enforce quoting conventions
    check_quotes(key, value, block_name)    
    
    return key, value


def _get_indent(line: str) -> int:
    """
    Get indentation level.

    Counts leading spaces and divides by 4 (1 level = 4 spaces).
    """
    if not line or not line[0].isspace():
        return 0
    spaces = len(line) - len(line.lstrip(' '))
    return spaces // 4


from ..src.menucmd import Menu

#Output type
class MenuDict(dict):
    """
    Dictionary of menu objects that can be additionally referenced by attributes:
    menu_dict['menu_id'] = menu_dict.menu_id
    """
    def __init__(self, menus: list):
        for menu in menus:
            self.__setattr__(menu._id, menu)

        self.menus = {menu._id: menu for menu in menus}

    def __getitem__(self, item) -> Menu:
        return self.menus[item]
    
    def get(self, item) -> Menu:
        if item not in self.menus:
            raise KeyError(f"{item} does not exist")
        return self.menus[item]
    
    def __iter__(self):
        return iter(self.menus)
    
    def __repr__(self):
        return f"<Menus {self.menus}>"
    
    def keys(self):
        return self.menus.keys()
    
    def values(self):
        return self.menus.values()
    
    def items(self):
        return self.menus.items()


from ..builtins.builtins import *
from .menu_dict import MenuDict
from .lines_to_dict import lines_to_dict
from .dict_to_objs import dict_to_objs

#==================================================================================

def build_menus(file: str) -> MenuDict:
    """Convert .mcmd file to attributable dict of menus"""

    #Get text
    with open(file) as f:
        lines = f.readlines()
    f.close()

    #Covert to dict
    struct_ = lines_to_dict(lines)

    #Convert to MenuDict
    menus = dict_to_objs(struct_)

    return menus



from menucmd import *
from menucmd import Bind as B
from menucmd.builtins import *
from menucmd.tests.testing_utils import *


def main():
    result = Menu.result
    kwargs = Menu.kwargs

    Menu.set_global_colors(
        colors=MenuColors(key=Colors.LIGHT_BLUE + Colors.BOLD), 
        exit_colors=ItemColors(key=Colors.RED)
    )

    main_menu = Menu(name = "Hub", exit_key = "x")
    lazy_menu = Menu(name = "Lazy Eval", exit_to = main_menu)
    builtin_menu = Menu(name = "Builtins", exit_to = main_menu, colors=MenuColors(key=Colors.CYAN  + Colors.BOLD))
    dynamic_menu = Menu(name = "Dynamic Menus", exit_to = main_menu, colors=MenuColors(key=Colors.CYAN + Colors.BOLD))

    menu_A = Menu(name="Menu_A", exit_to = lazy_menu, end_to = Menu.exit_to, colors=MenuColors(key=Colors.RED, message=Colors.FAINT))
    menu_B = Menu(name="Menu_B", exit_to = dynamic_menu, escape_to = Menu.exit_to, colors=MenuColors(key=Colors.RED, message=Colors.FAINT))

    main_menu.append(
        Item(key="e", message="Test Lazy Eval", funcs=[(lazy_menu, ())]),
        Item(key="b", message="Test Builtins", funcs=[(builtin_menu, ())]),
        Item(key="d", message="Test Dynamic Menus", funcs=[(dynamic_menu, ())]),
    )

    lazy_menu.append(
        Item(key='s', message='Square Number', funcs=[
            (input, ("Number: ")),
            (lambda x: x**2, B(float, result)),
            (print, (result)),
        ]),
        Item(key="c", message="Cards", funcs=[
            (shuffle_cards, ()),
            (pick_card, (B(int,B(input, "Choose Card (0-51): ")), result)),
            (print, (result)),
        ]),
        Item(key="m", message="menu composition", funcs=[
            (input, "string for menu_A: "), 
            (menu_A, result)
        ]),
        Item(key="i", message="iterator test", funcs=[
            (input, "list of 3 (separated by spaces)"), 
            (tuple, B(lambda x:x.split(), result)),
            (print, ("result call: ", result.expand())),
            (print, B(lambda x, y, z: x+y+z, result[-2].expand())),
        ]),
        Item(key="a", message="Attribute test",funcs=[
            (input, "integer 1: "), (int, result[-1].INT1),
            (input, "integer 2: "), (int, result[-1].INT2),
            (print, (result.INT1, result.INT2)),
            (print, ("INT1 : ", result.INT1[-1000000].INT1[314198].INT2.INT1)),
            (tuple, ((result.INT1, result.INT2),)),
            (print, ("expanded ", result[-1].EXP.expand())),
            (print, ("expanded (tagged)", result.EXP.expand()))
        ])
    )

    builtin_menu.append(
        Item(key="l", message="Edit List", funcs=[
            (input, ("List items (separated by spaces): ")),
            (edit_list, (B(lambda s: s.split(), result), kwargs(name = "Delete Items"))),
            (print, ("New string:\n", B(" ".join, result))),
        ]),
        Item(key="c", message="Choose Item", funcs=[
            (input, ("List items (separated by spaces): ")),
            (choose_item, (B(lambda s: s.split(), result), kwargs(name="Choose Item"))),
            (print, ("New string:\n", result)),
        ]),
        Item(key="v", message="Confirmation", funcs=[
            (yesno_ver, kwargs(name = "Delete system32?", exit_message = "no")),
            (f_switch(result, (f_escape, yesno_ver)), kwargs(name ="Are you really sure?", exit_message = "no")),
            (f_switch(result, (f_escape, yesno_ver)), kwargs(name = "Are you not sure?", exit_message = "no")),
            (f_switch(result, (yesno_ver, f_escape)), kwargs(name = "Are you sure you are not sure?", exit_message = "wtf?")),
            (print, B(lambda b: "crisis averted!" if b else "deleting...\nsystem32 cannot be deleted!", result))
        ]),
    )

    dynamic_menu.append(
        Item(key="x", message="test keyword function compositions", funcs=[
            (lambda: 0, ()),
            (menu_B, result),
        ]),
        Item(key="s", message="Menu.self usage: insert new item here", funcs=[
            (Menu.insert, (Menu.self, 0, 
                Item(key="*", message="Nice Job!", funcs=[(main_menu, "*")], 
                     colors=ItemColors(key=Colors.GREEN + Colors.BOLD)
                )
                # LIMITATION:
                # In order to have dynamic reference to other menus, the functions
                # must be defined in the same scope as the menus.  If you wanted to 
                # abstract the above out, it would therefore be impossible.
                # In the future there needs to be a global menu reference interface
                # as well as in the DSL!!
            )),
        ])
    )

    #==================================================================================
  
    menu_A.append(
        Item(key="m", message="reverse and print", funcs=[
            (lambda s: s[::-1], result),
            (print, result)
        ])
    )

    menu_B.append(
        Item(key="s", message="test escape_to", funcs=[
            (escape_on, (result, 3)),
            (print, result),
            (menu_B, B(lambda n: n+1, result[-2]))
        ])
    )

    #==================================================================================

    # Run
    main_menu()


#################################################################################################################

if __name__ == "__main__":
    main()


from dataclasses import dataclass, fields
from typing import Any, Optional


@dataclass 
class BaseData:
    def __iter__(self):
        for f in fields(self):
            yield getattr(self, f.name)


@dataclass
class Return(BaseData):
    val: Optional[Any] = None
    code: Optional[str] = None
    menu: Optional[str] = None
    err: Optional[Any] = None
    

@dataclass
class MenuEvalScope(BaseData):
    chain: list
    arg: Any
    tracked_attributes: dict



from dataclasses import dataclass
from .ansi_scheme import ItemColors
from typing import Optional, Any


@dataclass
class Item():
    key: str
    message: str
    funcs: list[tuple[Any, Any]]
    colors: Optional[ItemColors] = None



from ..src.menucmd import *
from macrolibs.typemacros import set_compliment, dict_compliment, type_compliment
from typing import Any, TypeVar

#----------------------------------------------------------------------------------------------------------------------
#In-Line Functions

def escape_on(x, value):
    """Returns Menu.escape if the two arguments are equal.
    Otherwise, return x."""
    return Menu.escape if x == value else x


def f_escape(*args, **kwargs) -> object:
    """Polymorphic in-line escape function."""   #terminal morphism in Hom(*,escape)
    return Menu.escape

def f_end(*args, **kwargs) -> None:
    """Polymorphic in-line None return (end)"""  #terminal morphism in Hom(*, None)
    return None


def f_switch(n: int | str | Any, funcs: list | tuple | dict) -> Bind.Wrapper:
    """Returns a lazy function of type (Any index -> function)"""
    return Bind(lambda b: maybe_arg(funcs[b]), n)



#----------------------------------------------------------------------------------------------------------------------
#Builtin Menus

Iterables = TypeVar("Iterables", list, tuple, dict, set)


def yesno_ver(yes = True, no = False, yes_message = "yes", **kwargs) -> bool | Any:
    """Simple yes/no verification returning bool by default
    Use yes and no tags to specify otherwise.
    """
    kwargs_: dict[str, Any] = ({"name":"Are you sure?", "exit_message":"cancel"} | kwargs |
               {"exit_to":lambda: no, "end_to":lambda: None})
    menu = Menu(**kwargs_)

    menu.append(Item(key="x", message=yes_message, funcs=[(lambda: yes, ())]))

    return menu()


def edit_list(entries: Iterables, display_as = lambda x:x, **kwargs) -> Iterables:
    """Delete items in a list/tuple/dict/set; returns updated list/tuple/dict/set"""

    kwargs_: dict[str, Any] = {"name":"Edit List"} | kwargs | {"exit_to":lambda: entries}
    menu = Menu(**kwargs_)

    if isinstance(entries, list | tuple):
        for n, entry in enumerate(entries):
            menu.append(Item(key=str(n+1), message=display_as(str(entry)), funcs=[
                (edit_list, (entries[:n] + entries[n+1:], Menu.kwargs(display_as=display_as, **kwargs)))
            ]))
    elif isinstance(entries, set):
        for n, entry in enumerate(entries):
            menu.append(Item(key=str(n+1), message=display_as(str(entry)), funcs =[
                (edit_list, (set_compliment(entries, {entry}), Menu.kwargs(display_as=display_as, **kwargs)))
            ]))
    elif isinstance(entries, dict):
        for n, (k, v) in enumerate(entries.items()):
            menu.append(Item(key=str(n+1), message=f"{k}:{display_as(str(v))}", funcs=[
                (edit_list, (dict_compliment(entries, {k:v}), Menu.kwargs(display_as=display_as, **kwargs)))
            ]))
    else:
        raise TypeError

    return type(entries)(menu()) #type: ignore


def choose_item(entries: list | tuple | dict | set, exit_val = None, display_as = lambda x:x, **kwargs) -> Any:
    """Pick and return an element from a list/tuple/dict/set.
    Returns (key, value) pair for dict.
    On exit key, return 'exit_val' (None by default)
    """
    kwargs_: dict[str, Any] = {"name": "Choose Item"} | kwargs | {"exit_to": lambda: exit_val}
    menu = Menu(**kwargs_)

    if isinstance(entries, list | tuple):
        for n, entry in enumerate(entries):
            menu.append(Item(key=str(n+1), message=display_as(str(entry)), funcs=[(Menu.id, entries[n])]))
    elif isinstance(entries, set):
        for n, entry in enumerate(entries):
            menu.append(Item(key=str(n+1), message=display_as(str(entry)), funcs=[(Menu.id, entry)]))
    elif isinstance(entries, dict):
        for n, (k, v) in enumerate(entries.items()):
            menu.append(Item(key=str(n+1), message=f"{k}:{display_as(str(v))}", funcs=[(Menu.id, ((k, v),))]))
    else:
        raise TypeError

    return menu()


def choose_items(entries: Iterables, display_as = lambda x:x, **kwargs) -> Iterables:
    """Pick and return mutiple elements from a list/tuple/dict/set."""
    return type_compliment(entries, edit_list(entries, display_as = display_as, **kwargs))


#----------------------------------------------------------------------------------------------------------------------
#Dynamic Menus

#WIP
def dynamic_wrapper(dyn_func, *args, **kwargs) -> Bind.Wrapper:
    """WIP
    Intended as a wrapper for arg_to (for now)

    Usage: Menu(arg_to = dynamic_wrapper(dyn_func, *args, **kwargs))

    Takes a func dyn_func: (menu_id, arg, *args, **kwargs) -> arg -> result[0]

    Example:
    def dyn_func(menu_id, arg, *items):
        menu_id.clear()
        menu_id.append(*items)
        return arg

    dyn_func must refer to the menu in its first argument, the menu argument for its second,
    and can take any additional *args/**kwargs.
    It is intended for arbitrary use of menu methods to dynamically change the menu on run.
    """
    return Bind(lambda menu_id, arg: dyn_func(menu_id, arg, *args, **kwargs), Menu.self)



from ..src.menucmd import Menu

def open_menu(menu: Menu, arg = None):
    """Open menu in func chain"""
    return Menu.__RUNMENU__(menu, arg)

def open_self(arg = None): 
    return Menu.__RUNSELF__(arg)

def to_menu(menu: Menu):
    """Open menu in Menu.*_to functions"""
    return lambda arg: open_menu(menu, arg)


from ..builtins.builtins import *
import inspect
from .menu_dict import MenuDict


# SHORTHANDS
result = Menu.result
kwargs = Menu.kwargs
self = Menu.self
B = Bind


def convert_static_attr_types(static_attrs: dict) -> None:
    """Converts Non-String types"""
    for key, val in static_attrs.items():
        if key not in { 'Colors', 'ExitColors' }:
            static_attrs[key] = eval(val)

    if static_attrs.get('Colors'):
        # Rename key for menu args
        static_attrs['colors'] = convert_colors(static_attrs['Colors'], "Menu")
        del static_attrs['Colors']

    if static_attrs.get('ExitColors'):
        # Rename key for menu args
        static_attrs['exit_colors'] = convert_colors(static_attrs['ExitColors'], "ExitColors")
        del static_attrs['ExitColors']


def convert_colors(colors: dict, block: str) -> MenuColors | ItemColors:
    """Evaluates ANSI color schemes"""
    scheme = {'Menu': MenuColors, 'Item': ItemColors, 'ExitColors': ItemColors}[block]
    return scheme(**dict([k, eval(v)] for k, v in colors.items()))


def retrieve_globals(target_globals):
    """Adds globals from where the .mcmd file is loaded."""
    caller_globals = inspect.stack()[3].frame.f_globals
    target_globals.update(caller_globals)


def cannonize_menu_ids(menus: MenuDict, target_globals):
    """Adds menu to global scope"""
    for menu_id, menu in menus.items():
        target_globals[menu_id] = menu


from macrolibs.typemacros import dict_intersect, dict_compliment
from .menu_dict import MenuDict
from ..src.item import Item
from ..builtins.builtins import *
from ..builtins.returns import *
from .utils import *
from ..src.menu_return import *
from ..src.colors import Colors as C


# Blank dict with only static attribute names
# Init Menu with these...
STATIC_ATTRS = {
    #"ID": '',
    "name": '',
    "clear_readout": '',
    "arg_to_first": '',
    "empty_message": '',
    "exit_message": '',
    "exit_key": '',
    "Colors": '',
    "ExitColors": '',
}

#==================================================================================

def dict_to_objs(struct_: dict) -> MenuDict:
    """
    Converts the dictionary structure to a MenuDict object.
    First initializes the menus with their static attributes, introduces the menus into the global
    scope referenced by their ids, and then appends their items and '*_to' callbacks after converting
    them to literals.
    """
    menus = []

    if not "Menu" in struct_:
        raise RuntimeError("'Menu' must be present in the outer struct")
    
    # Set Global Colors
    if global_colors := struct_.get("Colors"):
        Menu.set_global_colors(colors=convert_colors(global_colors, "Menu")) #type: ignore
    if global_colors := struct_.get("ExitColors"):
        Menu.set_global_colors(exit_colors=convert_colors(global_colors, "ExitColors"))

    # Initlialize all menus with only static attributes to a dict indexed by 'ID'
    for attrs in struct_["Menu"]:
        # Require id
        if not attrs.get('id'):
            raise KeyError("Menu must have an 'id' attribute")
        menu_id = attrs['id']
        
        # Compare all attributes defined in dsl script, and only feed into Menu 
        # if it intersects with the blank dict
        static_attrs = dict_intersect(attrs, STATIC_ATTRS)

        # Convert attrs types
        convert_static_attr_types(static_attrs)

        # Create Menu
        menu = Menu(**static_attrs)

        # Add id
        setattr(menu, "_id", menu_id)

        # Add to list
        menus.append(menu)

    # Convert list to attributable dict: {menu_id: Menu}
    menus = MenuDict(menus)

    # Add menu ids to global pointers and retrieve caller globals
    cannonize_menu_ids(menus, globals())
    retrieve_globals(globals())

    # Set all non-static attributes and convert 'ID' references to pointers
    for menu_id, menu, attrs in zip(menus.keys(), menus.values(), struct_["Menu"]):
        # Get all remaining attributes                            
        active_attributes = dict_compliment(attrs, STATIC_ATTRS)

        for name, attr in active_attributes.items():
            if name == "Item":
                _append_menu_items(menu, attr)
            else:
                # Adds the rest: arg_to, exit_to, etc
                setattr(menu, name, eval(attr))

        # Reapply menu updates
        menu.check_banned_self_refs()
        menu.apply_matching_keywords()
        menu.replace_self_references()
        menu.ch_exit()

    return menus


def _append_menu_items( menu: Menu, items: list[dict],):
    """
    Converts "Item" list into Items and appends to menu.
    [{key, message, funcs, Colors}, ...] -> Item
    """
    for item in items:
        if not item.get('key'):
            raise KeyError("Item must have 'key' attribute")

        funcs = [_parse_funcargs(funcargs) for funcargs in item['func']] if item.get('func') else []
        colors = convert_colors(item['Colors'], "Item") if item.get('Colors') else None  
        
        # Append Items
        menu.append(Item(
            key=eval(item["key"]), message=eval(item["message"]), 
            funcs=funcs, colors=colors
        )) 
        

def _parse_funcargs(funcargs: str) -> tuple[Any, tuple]:
    """"func(arg1, arg2,...)" -> (func, (arg1, arg2))"""
    # Extract function.  Works with (lambda:(...))
    func_args_split = funcargs.split('(')
    n_closed_par = 0

    # Split func and args
    for n, x in enumerate(func_args_split[::-1]):
        n_closed_par += x.count(')')
        if n_closed_par == n+1:
            func_str = "(".join(func_args_split[:-n-1])
            arg_str = "(" + "(".join(func_args_split[-n-1:])

            # Eval Func
            func = eval(func_str)
            
            # Eval Args
            try:
                args = () if arg_str.strip() == "()" \
                    else eval(arg_str.strip()[:-1] + ",)")
                
            # Menu.kwargs reminder
            except SyntaxError as e:
                if '=' in arg_str and '==' not in arg_str and ':=' not in arg_str:
                    raise SyntaxError(
                        f"Invalid keyword argument syntax in '{funcargs}'.\n"
                        f"All keyword arguments must be wrapped in 'Menu.kwargs()'.\n"
                    ) from e
                raise SyntaxError from e

            # Return tuple
            return (func, args)

    # Parenthesis Error
    raise SyntaxError(f"Unbalanced parenthesis in '{funcargs}'")



from dataclasses import dataclass, replace, asdict
from typing import Optional


@dataclass
class ItemColors:
    key: Optional[str] = None
    dash: Optional[str] = None
    message: Optional[str] = None

    def merge(self, other: Optional["ItemColors"]) -> "ItemColors":
        if other is None:
            return self
        return replace(self, **{k: v for k, v in asdict(other).items() if v is not None})


@dataclass
class MenuColors(ItemColors):
    name: Optional[str] = None
    empty_message: Optional[str] = None
    invalid_key: Optional[str] = None

    def merge(self, other: Optional["MenuColors | ItemColors"]) -> "MenuColors":
        if other is None:
            return self
        return replace(self, **{k: v for k, v in asdict(other).items() if v is not None})



import sys
from copy import copy
from dataclasses import dataclass
from typing import Any
from macrolibs.typemacros import tupler, dict_union, list_union
from macrolibs.typemacros import replace_value_nested, maybe_arg, maybe_type
from .result import Result
from .bind import Bind
from .colors import Colors as C
from .item import Item
from .ansi_scheme import *
from colorama import just_fix_windows_console

just_fix_windows_console()


#==================================================================================
# Menu Class
#==================================================================================

class _Expand(tuple):
    """Internal wrapper to mark an iterable for expansion as function arguments."""
    pass

class _Kwargs(dict):
    """Wrapper for kwargs in menufunction composition"""
    pass

@dataclass
class _RunMenu:
    """Return this to run a new menu in the call stack"""
    menu: "Menu"
    arg: Any

@dataclass 
class _RunSelf:
    """Return this to run self in the call stack"""
    arg: Any

class Menu:
    # Matching Keywords
    exit_to = object()  #Use as keyword for end_to to match exit_to:  Menu(end_to = Menu.exit_to)
    end_to = object()   #

    # Keyword Objects
    result = Result(-1)
    escape = object()
    self = object()
    kwargs = _Kwargs
    __END__ = type("Menu.__END__", (object,), {})  #Terminal Object
    id = lambda x:x                                #Identity morphism

    # Messages
    __RUNMENU__ = _RunMenu
    __RUNSELF__ = _RunSelf

    # Default Colors
    colors = MenuColors(
        name = C.BOLD,
        empty_message = C.LIGHT_RED + C.ITALIC,
        invalid_key = C.LIGHT_RED + C.ITALIC,
        key = C.BLINK + C.BOLD,
        dash = C.BOLD,
        message = C.ITALIC,
    )
    exit_colors = ItemColors()

    #==================================================================================

    def __init__(self,
                 name = "Choose Action",
                 invalid_key = "--*Invalid Key*--",
                 empty_message = "--*No Entries*--",
                 clear_readout = True,
                 arg_to = lambda x:x,
                 arg_to_first = True,
                 exit_to = lambda: Menu.__END__,
                 exit_key = "e",
                 exit_message = "exit",
                 colors = MenuColors(),
                 exit_colors = ItemColors(),
                 end_to = None,
                 escape_to = None,
                 #out_to = lambda x:x,
                ):
        # Menu
        self.name = name
        self.empty_message = empty_message
        self.invalid_key = invalid_key
        self.clear_readout = clear_readout

        # Arg
        self.arg_to = arg_to
        self.arg_to_first = arg_to_first

        # Exit
        self.exit_to = exit_to
        self.exit_key = exit_key
        self.exit_message =  exit_message
        
        # Colors
        self.colors = colors
        self.exit_colors = exit_colors

        #Init menu data
        self.menu_item_list: list[Item] = []         #de jure list of items
        self.menu_display_list = []                  #["[key]- message"]
        self.menu = {}                               #{"key":(function-arg chain)}

        #Define breakpoints, apply menu updates
        self.end_to = lambda x: Menu.__RUNSELF__(x) if end_to is None else end_to
        self.escape_to = lambda x: Menu.__RUNSELF__(x) if escape_to is None else escape_to
        self.check_banned_self_refs()
        self.apply_matching_keywords()
        self.replace_self_references()
        self.ch_exit()
        #TODO: group together?

    #==================================================================================
    # Stack Loop
    #==================================================================================

    def __call__(self, arg = None):
        current_menu = self
        current_arg = arg

        while True:
            menu_result = current_menu._run_menu(current_arg)

            if isinstance(menu_result, Menu.__RUNMENU__):
                current_menu = menu_result.menu
                current_arg = menu_result.arg
                continue
            elif isinstance(menu_result, Menu.__RUNSELF__):
                current_arg = menu_result.arg
                continue

            return menu_result

    #==================================================================================
    # Protocall
    #==================================================================================

    def _run_menu(self, arg = None):
        """Runs menu with 'arg' asking for user input.
        Selection runs the item's function chain where each nth function has
        access to Menu.result[0 -> n]; (n in [0,1,2...])

        Menu.result[0] == arg
        """
        # Merge Colors
        colors = Menu.colors.merge(self.colors)

        # Evaluate arg_to and add 0th result (Before selection)
        if self.arg_to_first:
            result = maybe_arg(self.arg_to)(arg)
            results = [result]

        # List state logic
        if not self.menu_display_list[:-1]:
            # Exit on empty menu
            empty_ansi = colors.empty_message
            print(f"{empty_ansi}{self.empty_message}\x1b[0m")
            selection = self.exit_key
        else:
            # Print Menu And Get Input
            name_ansi = colors.name
            show = f"\n{name_ansi}{self.name}\x1b[0m\n" + "\n".join(self.menu_display_list) + "\n"
            print(f"{show}\x1b[0m", end='')
            printed_lines = show.count('\n') + 1
            selection = input()

            # Clear previous menu
            if self.clear_readout:
                Menu.clear_lines(printed_lines)

        # Refresh Menu On Invalid Input
        if selection not in self.menu.keys():
            invalid_ansi = colors.invalid_key
            print(f"{invalid_ansi}{self.invalid_key}\x1b[0m")
            return Menu.__RUNSELF__(arg)

        # Select Item
        func_chain = self.menu[selection]

        # Exit menu if exit_key pressed
        if selection == self.exit_key:
            return maybe_arg(func_chain[0][0])(arg)

        # Evaluate arg_to and add 0th result (After selection)
        if not self.arg_to_first:
            result = maybe_arg(self.arg_to)(arg)
            results = [result]

        # Init Tracked Attributes
        tracked_attributes = {}

        # Evaluate Function Chain
        for pair in func_chain:
            # Get func/args pair
            func = pair[0]
            args = tupler(pair[1])

            # Replace Results and Separate args/kwargs
            func, args, kwargs = self.replace_keywords(func, args, results, tracked_attributes)

            # Evaluate Function
            result = Bind.lazy_eval(func, args, kwargs)

            # Restart Callstack 
            if result is Menu.__RUNMENU__:
                return result

            # Manual escape
            if result is Menu.escape:
                return maybe_arg(self.escape_to)(arg)

            # End Loop
            results.append(result)

        # Go to end_to if final value is None
        return result if result is not None else maybe_arg(self.end_to)(arg)

        #Maybe do this in the future for None returns:
        #maybe_arg(self.replace_keywords(self.end_to, (), results)[0])(arg)
        #
        #It doesn't make sense to have two refs to result[0] = arg
        #but built-in behaviour switching via Bind based on past results might be better than
        #using inline functions to switch between a result return and a None return


    def replace_keywords(self, func, args: tuple, results: list, tracked_attributes: dict) -> tuple:
        """Replaces all inline self references with the current menu.
        Replaces all Menu.result objects with function results from the chain.

        Returns a tuple in the form (func, args, kwargs).

        'args' may also include all kwargs distinguished by the Menu.kwargs wrapper.
        """
        # Replace self reference
        func = replace_value_nested(tupler(func), Menu.self, self)[0]
        args = replace_value_nested(tupler(args), Menu.self, self)

        # Replace Result Keywords by Attribute
        for tag, val in tracked_attributes.items():
            func = replace_value_nested(tupler(func), Result(0).__getattr__(tag), val)[0]
            args = replace_value_nested(tupler(args), Result(0).__getattr__(tag).expand(), maybe_type(_Expand, val))
            args = replace_value_nested(tupler(args), Result(0).__getattr__(tag), val)

        # Replace Result Keywords by Position
        N = len(results)
        for n in range(-N, N):
            # Detect if a result has an attribute and add first declaration to tracked attributes
            def track_attr(R, value):
                if R.__attr__ is not None and R.__attr__ not in tracked_attributes.keys():
                    tracked_attributes[R.__attr__] = value
                return value

            func = replace_value_nested(tupler(func), Result(n), results[n], callback= track_attr)[0]
            args = replace_value_nested(tupler(args), Result(n).expand(), maybe_type(_Expand, results[n]), callback= track_attr)
            args = replace_value_nested(tupler(args), Result(n), results[n], callback= track_attr)

        # Separate args/kwargs
        kwargs = dict_union(tupler(x for x in args if isinstance(x, Menu.kwargs)))
        args = tupler(x for x in args if not isinstance(x, Menu.kwargs))

        # 'Uninterprets' tupler for expanded results.  Allows inline notation for (*args) ~ result.expand()
        args = tuple(Menu.expand_in_place(args))

        return (func, args, kwargs)

    #==================================================================================
    # API
    #==================================================================================

    def append(self, *items: Item):
        """
        Append items to menu in the form:
        ('key', 'message', (func1, (*args1), func2, (*args2),...))
        --*{forces exit key to the end of the list}*--
        """
        Menu.check_item_type(*items)

        self.menu_display_list = self.menu_display_list[:-1]
        self.menu_item_list = self.menu_item_list[:-1]

        for item in items + (self.exit,):
            self.menu_item_list.append(item)
            self.update_menu_lists(item)


    def clear(self):
        """Clears all items from the menu"""
        self.menu_item_list, self.menu_display_list = [], []
        self.menu = {}
        self.update_menu_lists(self.exit)


    def insert(self, n: int, *items):
        """
        Insert items at position n in the form:
        ('key', 'message', (func1, (*args1), func2, (*args2),...))
        """
        Menu.check_item_type(*items)

        _data = self.menu_item_list[:-1][:n] + [*items] + self.menu_item_list[:-1][n:]
        self.clear()
        self.append(*_data)


    def delete(self, n: int, k: int = 1):
        """Delete k menu entries starting at position n"""
        _data = self.menu_item_list[:-1][:n] + self.menu_item_list[:-1][n+k:]
        self.clear()
        self.append(*_data)


    def __getitem__(self, idx):
        if isinstance(idx, slice):
            new_menu = copy(self)
            new_menu.clear()
            new_menu.append(*self.menu_item_list[:-1][idx])
            return new_menu
        return self.menu_item_list[idx]
    

    @classmethod
    def set_global_colors(
        cls, 
        colors: Optional[MenuColors] = None, 
        exit_colors: Optional[ItemColors] = None
    ):
        if colors:
            cls.colors = cls.colors.merge(colors)
        if exit_colors:
            cls.exit_colors = cls.exit_colors.merge(exit_colors)
    
    #==================================================================================
    # Updates
    #==================================================================================

    def update_menu_lists(self, item: Item):
        """Updates menu lists with new Item"""
        colors = Menu.colors.merge(self.colors).merge(item.colors)
        key_ansi = colors.key
        dash_ansi = colors.dash
        message_ansi = colors.message

        self.menu_display_list = list_union(
            self.menu_display_list, 
            [f"{key_ansi}[{item.key}]\x1b[0m{dash_ansi}-\x1b[0m {message_ansi}{item.message}\x1b[0m"]
        )
        self.menu[item.key] = item.funcs


    def ch_exit(self, exit_to = None, exit_key = None, exit_message = None) -> None:
        """Changes the properties of the exit key and appends it to the list"""
        self.exit_to = exit_to if exit_to else self.exit_to
        self.exit_key = exit_key if exit_key else self.exit_key
        self.exit_message = exit_message if exit_message else self.exit_message
        colors = Menu.exit_colors.merge(self.exit_colors)

        self.exit = Item(
            key=self.exit_key, message=self.exit_message, 
            funcs=[(self.exit_to, ())], colors=colors
        )
        self.append()


    def apply_matching_keywords(self):
        """Apply matching keywords for end_to and escape_to in order"""
        self.end_to = self.exit_to if self.end_to is Menu.exit_to else self.end_to
        self.escape_to = self.exit_to if self.escape_to is Menu.exit_to else self.escape_to
        self.escape_to = self.end_to if self.escape_to is Menu.end_to else self.escape_to


    def replace_self_references(self):
        """Replaces Menu.self with self in builtin menu functions"""
        menu_funcs = { 'arg_to': self.arg_to, 'end_to': self.end_to, 'escape_to': self.escape_to }
        for key in menu_funcs:
            x_to = menu_funcs[key]
            replaced = self if x_to is Menu.self \
                else replace_value_nested(tupler(x_to), Menu.self, self)[0]
            setattr(self, key, replaced)

    #==================================================================================
    # Util
    #==================================================================================

    @staticmethod
    def expand_in_place(A: tuple | list):
        """Recursively expand all marked tuples/lists in a data structure
        (a, [b, expanded((c, d))]) -> (a, [b, c, d])
        """
        new = []
        for x in A:
            # Append all elements if marked as 'expanded'
            if isinstance(x, _Expand):
                new += list(x)
            # Run in nesting (recursion)
            elif isinstance(x, tuple | list):
                new += [Menu.expand_in_place(x)]
            # Do nothing
            else:
                new += [x]
        return type(A)(new)


    @staticmethod
    def clear_lines(printed_lines: int) -> None:
        """Clears previous menu from terminal"""
        if printed_lines <= 0:
            return
        #print(f"\x1b[{printed_lines}A\x1b[J", end='', flush=True)
        sys.stdout.write(f"\x1b[{printed_lines}A\x1b[J")
        sys.stdout.flush()

    #==================================================================================
    # Errors
    #==================================================================================

    @staticmethod
    def check_item_type(*items):
        for item in items:
            if not isinstance(item, Item):
                raise ValueError(f"{item} is not a valid Item")
            

    def check_banned_self_refs(self):
        banned_methods = { 'exit_to': self.exit_to, 'arg_to': self.arg_to  }
        for method in banned_methods:
            if banned_methods[method] is Menu.self:
                raise RecursionError(f"Menu \"{self.name}\" {method} cannot be Menu.self")
        


from macrolibs.typemacros import tupler



class Bind():
    """
    Used for delayed evaluation.
    Creates a callable list of function / argument pairs in the form:
    [<function>, (args), {kwargs}]  ->  function(*args, **kwargs)

    - Calling a Bind object with additional args/kwargs will append (curry) them to its initial args/kwargs.

      Bind(func, args1, kwargs1)(args2, kwargs2) -> func(*args1, *args2, **kwargs1, **kwargs2)

    - Use Bind(func, *args, **kwargs).fix() to toggle currying.
    """
    class Wrapper(list):
        def __init__(self, data = ()):
            super().__init__(data)
            self.fixed = False

        def __call__(self, *args, **kwargs):
            if self.fixed:
                return Bind.lazy_eval(self[0], self[1], self[2])
            else:
                return Bind.lazy_eval(self[0], self[1] + args, self[2] | kwargs)

        def fix(self):
            self.fixed = not self.fixed
            return self

    def __new__(cls, func, *args, **kwargs) -> Wrapper:
        return cls.Wrapper([func, args, kwargs])


    @staticmethod
    def lazy_eval(func, args = (), kwargs = {}):
        """Depth-first evaluation of nested function/argument bindings."""
        func = func() if isinstance(func, Bind.Wrapper) else func
        args = tupler(arg() if isinstance(arg, Bind.Wrapper) else arg for arg in tupler(args))
        kwargs = {k: v() if isinstance(v, Bind.Wrapper) else v for k, v in kwargs.items()}

        return func(*args, **kwargs)

"""""
Do this to evaluate Binds recursively through ALL data structures, not just immediate Bind nestings.
This might not be the best usage because the user cannot use Bind objects in a data structure without evaluating
everything at the same time.  Adding a keyword argument to change the desired behaviour might be the right course
but for menucmd, it does not matter.

    
    def lazy_eval(func, args = (), kwargs = {}):
        func = func() if isinstance(func, Bind.Wrapper) else func

        def r_eval(data):
            if isinstance(data, Bind.Wrapper):
                return data()
            elif isinstance(data, list | tuple):
                return type(data)(r_eval(x) for x in data)
            else:
                return data

        args = tupler(r_eval(arg) for arg in tupler(args))
        kwargs = {k:r_eval(v) for k, v in kwargs.items()}
    
        return func(*args, **kwargs)
"""""



from .src.menucmd import Menu
from .src.item import Item
from .src.bind import Bind
from .src.result import Result

from .src.ansi_scheme import MenuColors, ItemColors
from .src.colors import Colors

__all__ = [
	'Menu', 'Item', 'Bind', 'Result',
	'MenuColors', 'ItemColors', 'Colors',
]